@@ -15,12 +15,14 @@ module RSpec
         # @api private
         # @return [Boolean]
         def matches?(actual)
+          actual = actual.to_hash if convert_to_hash?(actual)
           perform_match(actual) { |v| v }
         end
 
         # @api private
         # @return [Boolean]
         def does_not_match?(actual)
+          actual = actual.to_hash if convert_to_hash?(actual)
           perform_match(actual) { |v| !v }
         end
 
@@ -137,6 +139,10 @@ module RSpec
             actual.include?(str) && lines.none? { |line| line == str }
           end
         end
+
+        def convert_to_hash?(obj)
+          !(::Hash === obj) && obj.respond_to?(:to_hash)
+        end
       end
     end
   end
